@@ -31,6 +31,7 @@ using System;
 using System.IO;
 using System.Net;
 using System.Net.Sockets;
+using System.Security.Cryptography.X509Certificates;
 using Mono.Unix;
 
 namespace Mono.WebServer
@@ -39,7 +40,7 @@ namespace Mono.WebServer
 	// ModMonoWebSource: Provides methods to get objects and types specific
 	// to mod_mono.
 	//
-	public class ModMonoWebSource: IWebSource, IDisposable
+	public class ModMonoWebSource: WebSource
 	{
 		string filename;
 		bool file_bound;
@@ -94,7 +95,7 @@ namespace Mono.WebServer
 			return result;
 		}
 
-		public virtual Socket CreateSocket ()
+		public override Socket CreateSocket ()
 		{
 			if (filename == null)
 				throw new InvalidOperationException ("filename not set");
@@ -106,7 +107,7 @@ namespace Mono.WebServer
 					conn.Connect (ep);
 					conn.Close ();
 					throw new InvalidOperationException ("There's already a server listening on " + filename);
-				} catch (SocketException se) {
+				} catch (SocketException) {
 				}
 				File.Delete (filename);
 			}
@@ -117,27 +118,22 @@ namespace Mono.WebServer
 			return listen_socket;
 		}
  
-		public IWorker CreateWorker (Socket client, ApplicationServer server)
+		public override Worker CreateWorker (Socket client, ApplicationServer server)
 		{
 			return new ModMonoWorker (client, server);
 		}
 		
-		public Type GetApplicationHostType ()
+		public override Type GetApplicationHostType ()
 		{
 			return typeof (ModMonoApplicationHost);
 		}
 		
-		public IRequestBroker CreateRequestBroker ()
+		public override IRequestBroker CreateRequestBroker ()
 		{
 			return new ModMonoRequestBroker ();
 		}
 
-		public void Dispose ()
-		{
-			Dispose (true);
-		}
-
-		protected virtual void Dispose (bool expl)
+		protected override void Dispose (bool expl)
 		{
 			if (filename != null) {
 				string f = filename;
@@ -192,6 +188,21 @@ namespace Mono.WebServer
 	//
 	public class ModMonoApplicationHost : BaseApplicationHost
 	{
+		private byte[] FromPEM (string pem) 
+		{
+			int start = pem.IndexOf ("-----BEGIN CERTIFICATE-----");
+			if (start < 0)
+				return null;
+
+			start += 27; // 27 being the -----BEGIN CERTIFICATE----- length
+			int end = pem.IndexOf ("-----END CERTIFICATE-----", start);
+			if (end < start)
+				return null;
+				
+			string base64 = pem.Substring (start, (end - start));
+			return Convert.FromBase64String (base64);
+		}
+
 		public void ProcessRequest (int reqId, string verb, string queryString, string path,
 					string protocol, string localAddress, int serverPort, string remoteAddress,
 					int remotePort, string remoteName, string [] headers, string [] headerValues)
@@ -200,6 +211,38 @@ namespace Mono.WebServer
 			ModMonoWorkerRequest mwr = new ModMonoWorkerRequest (reqId, broker, this, verb, path, queryString,
 									protocol, localAddress, serverPort, remoteAddress,
 									remotePort, remoteName, headers, headerValues);
+			if (mwr.IsSecure ()) {
+				// note: we're only setting what we use (and not the whole lot)
+				mwr.AddServerVariable ("CERT_KEYSIZE", broker.GetServerVariable (reqId, "SSL_CIPHER_USEKEYSIZE"));
+				mwr.AddServerVariable ("CERT_SECRETKEYSIZE", broker.GetServerVariable (reqId, "SSL_CIPHER_ALGKEYSIZE"));
+
+				string pem_cert = broker.GetServerVariable (reqId, "SSL_CLIENT_CERT");
+				// 52 is the minimal PEM size for certificate header/footer
+				if ((pem_cert != null) && (pem_cert.Length > 52)) {
+					byte[] certBytes = FromPEM (pem_cert);
+					mwr.SetClientCertificate (certBytes);
+
+					// check client certificate validity with Apache and/or Mono
+					if (mwr.IsClientCertificateValid (certBytes)) {
+						// client cert present (bit0 = 1) and valid (bit1 = 0)
+						mwr.AddServerVariable ("CERT_FLAGS", "1");
+					} else {
+						// client cert present (bit0 = 1) but invalid (bit1 = 1)
+						mwr.AddServerVariable ("CERT_FLAGS", "3");
+					}
+				} else {
+					mwr.AddServerVariable ("CERT_FLAGS", "0");
+				}
+
+				pem_cert = broker.GetServerVariable (reqId, "SSL_SERVER_CERT");
+				// 52 is the minimal PEM size for certificate header/footer
+				if ((pem_cert != null) && (pem_cert.Length > 52)) {
+					byte[] certBytes = FromPEM (pem_cert);
+					X509Certificate cert = new X509Certificate (certBytes);
+					mwr.AddServerVariable ("CERT_SERVER_ISSUER", cert.GetIssuerName ());
+					mwr.AddServerVariable ("CERT_SERVER_SUBJECT", cert.GetName ());
+				}
+			}
 
 			ProcessRequest (mwr);
 		}
@@ -209,7 +252,7 @@ namespace Mono.WebServer
 	// ModMonoWorker: The worker that do the initial processing of mod_mono
 	// requests.
 	//
-	internal class ModMonoWorker: IWorker
+	internal class ModMonoWorker: Worker
 	{
 		ApplicationServer server;
 		public LingeringNetworkStream Stream;
@@ -226,12 +269,19 @@ namespace Mono.WebServer
 		int requestId = -1;
 		ModMonoRequestBroker broker = null;
 			
-		public void Run (object state)
+		public override void Run (object state)
 		{
 			try {
 				InnerRun (state);
+			} catch (FileNotFoundException fnf) {
+				// We print this one, as it might be a sign of a bad deployment
+				// once we require the .exe and Mono.WebServer in bin or the GAC.
+				Console.Error.WriteLine (fnf);
+			} catch (IOException) {
+				// This is ok (including EndOfStreamException)
 			} catch (Exception e) {
-				//Console.WriteLine ("In ModMonoWorker.Run: {0}", e.Message);
+				Console.Error.WriteLine (e);
+			} finally {
 				try {
 					// Closing is enough for mod_mono. the module will return a 50x
 					Stream.Close ();
@@ -277,10 +327,11 @@ namespace Mono.WebServer
 			}
 
 
-			Console.Error.WriteLine ("final_pdir: {0} final_vdir: {1}", final_pdir, final_vdir);
+			//Console.Error.WriteLine ("final_pdir: {0} final_vdir: {1}", final_pdir, final_vdir);
 			vapp = server.GetApplicationForPath (vhost, port, virt, false);
 			if (vapp == null) {
-				final_pdir = "file://" + final_pdir;
+				// Don't know why this breaks mod-mono-server2.exe, but not mod-mono-server.exe
+				//final_pdir = "file://" + final_pdir;
 				if (final_vdir [0] != '/')
 					final_vdir = "/" + final_vdir;
 				server.AddApplication (vhost, port, final_vdir, final_pdir);
@@ -372,17 +423,17 @@ namespace Mono.WebServer
 								modRequest.GetAllHeaderValues());
 		}
 
-		public int Read (byte[] buffer, int position, int size)
+		public override int Read (byte[] buffer, int position, int size)
 		{
 			return modRequest.GetClientBlock (buffer, position, size);
 		}
 		
-		public void Write (byte[] buffer, int position, int size)
+		public override void Write (byte[] buffer, int position, int size)
 		{
 			modRequest.SendResponseFromMemory (buffer, position, size);
 		}
 		
-		public void Close ()
+		public override void Close ()
 		{
 			if (closed)
 				return;
@@ -399,12 +450,12 @@ namespace Mono.WebServer
 			} catch {}
 		}
 		
-		public void Flush ()
+		public override void Flush ()
 		{
 			//modRequest.Flush (); No-op
 		}
 
-		public bool IsConnected ()
+		public override bool IsConnected ()
 		{
 			return modRequest.IsConnected ();
 		}
@@ -430,3 +481,4 @@ namespace Mono.WebServer
 		}
 	}	
 }
+
